//$Header: /deegreerepository/deegree/src/org/deegree/ogcwebservices/wfs/WFService.java,v 1.46 2007/03/14 14:43:44 mschneider Exp $
/*----------------    FILE HEADER  ------------------------------------------

 This file is part of deegree.
 Copyright (C) 2001-2008 by:
 EXSE, Department of Geography, University of Bonn
 http://www.giub.uni-bonn.de/deegree/
 lat/lon GmbH
 http://www.lat-lon.de

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 Contact:

 Andreas Poth
 lat/lon GmbH
 Aennchenstr. 19
 53115 Bonn
 Germany
 E-Mail: poth@lat-lon.de

 Prof. Dr. Klaus Greve
 Department of Geography
 University of Bonn
 Meckenheimer Allee 166
 53115 Bonn
 Germany
 E-Mail: greve@giub.uni-bonn.de

 ---------------------------------------------------------------------------*/
package org.deegree.ogcwebservices.wfs;

import java.util.Map;

import org.deegree.datatypes.QualifiedName;
import org.deegree.framework.log.ILogger;
import org.deegree.framework.log.LoggerFactory;
import org.deegree.framework.trigger.TriggerProvider;
import org.deegree.io.datastore.LockManager;
import org.deegree.io.datastore.schema.MappedFeatureType;
import org.deegree.ogcwebservices.OGCWebService;
import org.deegree.ogcwebservices.OGCWebServiceException;
import org.deegree.ogcwebservices.OGCWebServiceRequest;
import org.deegree.ogcwebservices.wfs.capabilities.WFSCapabilities;
import org.deegree.ogcwebservices.wfs.configuration.WFSConfiguration;
import org.deegree.ogcwebservices.wfs.operation.DescribeFeatureType;
import org.deegree.ogcwebservices.wfs.operation.GetFeature;
import org.deegree.ogcwebservices.wfs.operation.GetFeatureWithLock;
import org.deegree.ogcwebservices.wfs.operation.LockFeature;
import org.deegree.ogcwebservices.wfs.operation.WFSGetCapabilities;
import org.deegree.ogcwebservices.wfs.operation.transaction.Transaction;

/**
 * This class implements access to the methods defined in the OGC WFS 1.1.0 specification.
 * 
 * @author <a href="mailto:schneider@lat-lon.de">Markus Schneider</a>
 * @author <a href="mailto:poth@lat-lon.de">Andreas Poth</a>
 * 
 * @author last edited by: $Author: apoth $
 * 
 * @version $Revision: 9345 $, $Date: 2007-12-27 17:22:25 +0100 (Do, 27 Dez 2007) $
 * 
 * @see OGCWebService
 */
public class WFService implements OGCWebService {

    private static final ILogger LOG = LoggerFactory.getLogger( WFService.class );

    /** Only OGC standard version currently implemented by this service. */
    public static final String VERSION = "1.1.0";

    private static final TriggerProvider TP = TriggerProvider.create( WFService.class );

    private WFSConfiguration configuration;

    // shared instance that handles all GetFeature requests to this service
    private GetFeatureHandler getFeatureHandler;

    // shared instance that handles all DescribeFeatureType requests to this service
    private DescribeFeatureTypeHandler describeFTHandler;

    // shared instance that handles all LockFeature requests to this service
    private LockFeatureHandler lockFeatureHandler;

    /**
     * Creates a new instance of <code>WFService</code> with the given configuration.
     * 
     * @param configuration
     * @throws OGCWebServiceException
     */
    WFService( WFSConfiguration configuration ) throws OGCWebServiceException {
        this.configuration = configuration;
        this.getFeatureHandler = new GetFeatureHandler( this );
        this.describeFTHandler = new DescribeFeatureTypeHandler( this );
        this.lockFeatureHandler = new LockFeatureHandler( this );
    }

    /**
     * Returns the capabilities of the <code>WFService</code>.
     * 
     * @return the capabilities, this is actually a <code>WFSConfiguration</code> instance
     */
    public WFSCapabilities getCapabilities() {
        return this.configuration;
    }

    /**
     * Performs the handling of the passed OGCWebServiceEvent directly and returns the result to the
     * calling class/ method.
     * 
     * @param request
     *            WFS request to perform
     * 
     * @throws OGCWebServiceException
     */
    public Object doService( OGCWebServiceRequest request )
                            throws OGCWebServiceException {

        long ts = System.currentTimeMillis();
        request = (OGCWebServiceRequest) TP.doPreTrigger( this, request )[0];

        Object response = null;
        char rtype = 'C';
        if ( request instanceof WFSGetCapabilities ) {
            // TODO implement partial responses (if only some sections are requested)
            response = this.configuration;
            rtype = 'C';
        } else if ( request instanceof GetFeatureWithLock ) {
            response = this.lockFeatureHandler.handleRequest( (GetFeatureWithLock) request );
            rtype = 'W';
        } else if ( request instanceof GetFeature ) {
            response = this.getFeatureHandler.handleRequest( (GetFeature) request );
            rtype = 'F';
        } else if ( request instanceof DescribeFeatureType ) {
            response = this.describeFTHandler.handleRequest( (DescribeFeatureType) request );
            rtype = 'D';
        } else if ( request instanceof Transaction ) {
            TransactionHandler handler = new TransactionHandler( this, (Transaction) request );
            response = handler.handleRequest();
            rtype = 'T';
        } else if ( request instanceof LockFeature ) {
            response = this.lockFeatureHandler.handleRequest( (LockFeature) request );
            rtype = 'L';
        } else {
            String msg = "Unknown request type: " + request.getClass().getName();
            throw new OGCWebServiceException( getClass().getName(), msg );
        }
        if ( LOG.getLevel() == ILogger.LOG_DEBUG ) {
            LOG.logDebug( "" + LockManager.getInstance() );
        }

        Object o = TP.doPostTrigger( this, response )[0];
        LOG.logDebug( "WFS processing time for request type " + rtype + ": ",
                      Long.toString( ( System.currentTimeMillis() - ts ) ) );
        return o;
    }

    /**
     * Returns a clone of the <code>WFService</code> instance.
     * <p>
     * Note that the configuration of the new service will refer to the same instance.
     */
    @Override
    public Object clone() {

        WFService clone = null;
        try {
            clone = new WFService( configuration );
        } catch ( OGCWebServiceException e ) {
            // can never happen
        }
        return clone;
    }

    /**
     * Returns the <code>MappedFeatureType</code> with the given name.
     * 
     * @param name
     *            name of the feature type
     * @return the mapped feature type with the given name, or null if it is not known to this
     *         WFService instance
     */
    public MappedFeatureType getMappedFeatureType( QualifiedName name ) {
        return this.configuration.getMappedFeatureTypes().get( name );
    }

    /**
     * Returns a <code>Map</code> of the feature types that this WFS serves.
     * 
     * @return keys: feature type names, values: mapped feature types
     */
    public Map<QualifiedName, MappedFeatureType> getMappedFeatureTypes() {
        return this.configuration.getMappedFeatureTypes();
    }
}