//$HeadURL: svn+ssh://rbezema@svn.wald.intevation.org/deegree/base/branches/2.2_testing/src/org/deegree/ogcwebservices/wfs/operation/WFSGetCapabilities.java $
/*----------------    FILE HEADER  ------------------------------------------

 This file is part of deegree.
 Copyright (C) 2001-2008 by:
 EXSE, Department of Geography, University of Bonn
 http://www.giub.uni-bonn.de/deegree/
 lat/lon GmbH
 http://www.lat-lon.de

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 Contact:

 Andreas Poth
 lat/lon GmbH
 Aennchenstraße 19
 53177 Bonn
 Germany
 E-Mail: poth@lat-lon.de

 Prof. Dr. Klaus Greve
 Department of Geography
 University of Bonn
 Meckenheimer Allee 166
 53115 Bonn
 Germany
 E-Mail: greve@giub.uni-bonn.de
 
 ---------------------------------------------------------------------------*/
package org.deegree.ogcwebservices.wfs.operation;

import java.util.Map;

import org.deegree.framework.util.KVP2Map;
import org.deegree.ogcwebservices.InvalidParameterValueException;
import org.deegree.ogcwebservices.MissingParameterValueException;
import org.deegree.ogcwebservices.OGCWebServiceException;
import org.deegree.ogcwebservices.getcapabilities.GetCapabilities;
import org.deegree.ogcwebservices.wfs.WFService;
import org.w3c.dom.Element;

/**
 * Represents a GetCapabilities request to a web feature service.
 * <p>
 * The GetCapabilities request is used to query a capabilities document from a web feature service.
 * 
 * @author <a href="mailto:poth@lat-lon.de">Andreas Poth </a>
 * @author <a href="mailto:schneider@lat-lon.de">Markus Schneider </a>
 * @author last edited by: $Author: apoth $
 * 
 * @version $Revision: 9345 $, $Date: 2007-12-27 17:22:25 +0100 (Do, 27 Dez 2007) $
 */
public class WFSGetCapabilities extends GetCapabilities {

    private static final long serialVersionUID = 3581485156939911513L;

    /**
     * Creates a new <code>WFSGetCapabilities</code> instance. with a wfs version of {@link WFService.VERSION}.
     * 
     * @param id
     *            request identifier
     * @param updateSeq
     * @param acceptVersions
     * @param sections
     * @param acceptFormats
     * @param vendoreSpec
     */
    WFSGetCapabilities( String id, String updateSeq, String[] acceptVersions, String[] sections,
                        String[] acceptFormats, Map<String, String> vendoreSpec ) {
        this( id, WFService.VERSION, updateSeq, acceptVersions, sections, acceptFormats, vendoreSpec );
    }

    /**
     * Creates a new <code>WFSGetCapabilities</code> instance.
     * 
     * @param id
     *            request identifier
     * @param version
     *            the version of the request. (e.g 1.0.0 or 1.1.0)
     * @param updateSeq
     * @param acceptVersions
     * @param sections
     * @param acceptFormats
     * @param vendoreSpec
     */
    WFSGetCapabilities( String id, String version, String updateSeq, String[] acceptVersions, String[] sections,
                        String[] acceptFormats, Map<String, String> vendoreSpec ) {
        super( id, version, updateSeq, acceptVersions, sections, acceptFormats, vendoreSpec );
    }

    /**
     * Creates a <code>WFSGetCapabilities</code> instance from a document that contains the DOM representation of the
     * request.
     * 
     * @param id
     * @param root
     *            element that contains the DOM representation of the request
     * @return transaction instance
     * @throws OGCWebServiceException
     */
    public static WFSGetCapabilities create( String id, Element root )
                                                                      throws OGCWebServiceException {
        WFSGetCapabilitiesDocument doc = new WFSGetCapabilitiesDocument();
        doc.setRootElement( root );
        WFSGetCapabilities request;
        try {
            request = doc.parse( id );
        } catch ( Exception e ) {
            throw new OGCWebServiceException( "WFSGetCapabilities", e.getMessage() );
        }
        return request;
    }

    /**
     * Creates a new <code>WFSGetCapabilities</code> instance from the given key-value pair encoded request.
     * 
     * @param id
     *            request identifier
     * @param request
     * @return new <code>WFSGetCapabilities</code> request
     * @throws InvalidParameterValueException
     * @throws MissingParameterValueException
     */
    public static WFSGetCapabilities create( String id, String request )
                                                                        throws InvalidParameterValueException,
                                                                        MissingParameterValueException {
        Map<String, String> map = KVP2Map.toMap( request );
        map.put( "ID", id );
        return create( map );
    }

    /**
     * Creates a new <code>WFSGetCapabilities</code> request from the given map.
     * 
     * @param request
     * @return new <code>WFSGetCapabilities</code> request
     * @throws InvalidParameterValueException
     * @throws MissingParameterValueException
     */
    public static WFSGetCapabilities create( Map<String, String> request )
                                                                          throws InvalidParameterValueException,
                                                                          MissingParameterValueException {

        String service = getRequiredParam( "SERVICE", request );
        if ( !service.equals( "WFS" ) ) {
            throw new InvalidParameterValueException( "WFSGetCapabilities", "Parameter 'service' must be 'WFS'." );
        }
        String version = request.get( "VERSION" );
        String[] acceptVersions = getParamValues( "ACCEPTVERSIONS", request, WFService.VERSION );
        String[] sections = getParamValues( "SECTIONS", request, "" );
        String updateSequence = getParam( "UPDATESEQUENCE", request, "" );
        String[] acceptFormats = getParamValues( "ACCEPTFORMATS", request, "text/xml" );

        // TODO generate unique request id
        String id = null;
        if ( version == null || "".equals( version.trim() ) ) {
            return new WFSGetCapabilities( id, updateSequence, acceptVersions, sections, acceptFormats, request );
        }
        return new WFSGetCapabilities( id, version, updateSequence, acceptVersions, sections, acceptFormats, request );
    }

    /**
     * Returns the service name (WFS).
     * 
     * @return the service name (WFS).
     */
    public String getServiceName() {
        return "WFS";
    }
}
